/*****************************************************************************/
/*                                                                           */
/*      Version:  1.00   Date: 19/04/96   File: main-trial.c                 */
/* Last Version:                          File:                              */
/* Changes:                                                                  */
/* 19/04/96 Created                                                          */
/*                                                                           */
/* Purpose:                                                                  */
/*                                                                           */
/*                                                                           */
/* Author:  Thomas Stuetzle                                                  */
/*                                                                           */
/*****************************************************************************/
/*                                                                           */
/*===========================================================================*/

#include <stdio.h>
#include <math.h>
#include <limits.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <getopt.h>

#ifdef URAND
  #include "urand.h"
#endif

#include "timer.h"
#include "stat.h"

#define LINE_BUF_LEN  512
#define PROG_ID_STR "Basic input routine and local search for QAP, V0.1\n"
#define CALL_SYNTAX_STR "lsqap <instance>\n" 
#define VERBOSE(x) x
#define VERYVERBOSE(x)
 
#define XOR(x,y) ((x && !y) || (!x && y))
#define MIN(x,y)        ((x)<=(y)?(x):(y))
#define MAX(x,y)        ((x)>=(y)?(x):(y))

#define FALSE 0
#define TRUE  1

/* --- variables for main program ---------------------------------------- */

/* constants for a pseudo-random number generator, details see Numerical Recipes in C book */

#define IA 16807
#define IM 2147483647
#define AM (1.0/IM)
#define IQ 127773
#define IR 2836

long int seed; /* seed for the random number generator */

long int opt;

long int **d;  /* first matrix read from input file, typically distance matrix */
long int **f;  /* second matrix read from input file, typically flow matrix */

long int  *p;  /* array containing the current solution. if d is the distance matrix 
		  and f the flow matrix, p[i] is the item assigned to location i
	       */

long int  n;                    /* instance size of the QAP instance */
long int  best_known;           /* best objective function value of QAP instance */

long int  best_found;           /* best solution found in local search */

long int null_diagonal_flag  = FALSE;  /* at least one matrix has zero diagonal: TRUE */
long int d_symmetric_flag    = FALSE;  /* if first (d) matrix is symmetric: TRUE */
long int f_symmetric_flag    = FALSE;  /* if second (f) matrix is symmetric: TRUE */
long int make_symmetric_flag = FALSE;  /* convert asymmetric instance into symmetric 
					  instance: TRUE
				       */

long int first_it_flag;         /* first iteration of local search: yes / no */

FILE    *output_file;           /* output file for developement of solution quality */
FILE    *output_summary_file;   /* output file for developement of solution quality */
char name_buf[LINE_BUF_LEN];    /* input file name */

long int   **move_values;       /* move values in best improvement local search */
long int   **tabu_values;       /* entries give iteration up to which an attribute is forbidden */ 

long int   rchosen, schosen;    /* locations involved in move */
long int   tabu_search_length;  /* how many iterations the tabu search is run, is a multiple of instance size */
long int   tabu_list_length;    
long int   iterations_before_aspiration;  /* robust tabu search specific, multiple of instance size */
long int   aspirating_iteration;
long int   max_trials;          /* max. number of repetitions of robust tabu search */
long int   trials;              /* Number trials */

double     time_limit;          /* time limit */

long int  *best_in_try;       /* store some summary data */
long int  *best_found_at;     /* store some summary data */
double    *t_best_found;      /* store some summary data */

long int  iteration_best_found;      /* iteration at which best solution is found */
double    time_best_found = 0.0;    /* stores the time at which best solution is found 
					in each trial */

double ran01( long *idum ) {
/* 
      FUNCTION:      returns a pseudo-random number
      INPUT:         a pointer to the seed variable 
      OUTPUT:        a pseudo-random number uniformly distributed in [0,1]
      (SIDE)EFFECTS: changes the value of seed
*/
  long k;
  double ans;

  k =(*idum)/IQ;
  *idum = IA * (*idum - k * IQ) - IR * k;
  if (*idum < 0 ) *idum += IM;
  ans = AM * (*idum);
  return ans;
}

void read_problem_size( FILE *input ) {
/* 
      FUNCTION:      read the dimension of the QAP instanced/
      INPUT:         a pointer to the input file 
      OUTPUT:        none
      (SIDE)EFFECTS: assigns n the instance dimension
*/
  if( fscanf(input, "%ld", &n) <= 0) {
    fprintf(stderr, "error reading qap size value in data file\n");
    exit(1);	
  }
  VERYVERBOSE ( printf("QAP instance size %ld\n",n); )
}

void read_best_known_value( FILE *input ) {
/* 
      FUNCTION:      read the best known objective function value of the QAP instance
      INPUT:         a pointer to the input file 
      OUTPUT:        none
      (SIDE)EFFECTS: assigns best_known the read value
*/
  if( fscanf(input, "%ld ", &best_known) < 0 ) {
    fprintf(stderr, "error reading best known solution in data file\n");
    exit(1);	
  }
  VERYVERBOSE ( printf(" Best known solution %ld\n",best_known); )
}

long int ** read_matrix( FILE *input, long int size )
/*    
      FUNCTION:      read in a QAP instance matrix
      INPUT:         Pointer to input file, size of QAP instance
      OUTPUT:        pointer to matrix, has to be freed before program stops
      (SIDE)EFFECTS: allocates a memory of appropriate size for the matrix
*/
{
  long int     i, j;
  long int     **matrix;

  if((matrix = malloc(sizeof(long int) * size * size +
		      sizeof(long int *) * size	 )) == NULL){
    printf("Out of memory, exit.");
    exit(1);
  }
  for ( i = 0 ; i < size ; i++ ) {
    matrix[i] = (long int *)(matrix + size) + i*size;
    for ( j = 0  ; j < size ; j++ ) {
      if( fscanf(input, "%ld", &matrix[i][j]) < 0) {
	fprintf(stderr, "error reading (%ld, %ld) qap_distance in data file\n", i, j);
	exit(1);	
      }
    }
  }
  return matrix;
}

long int check_null_diagonal ( long int **matrix, long int size ) {
/* 
      FUNCTION:      check whether the Matrix matrix has a zero diagonal
      INPUT:         pointer to the matrix
      OUTPUT:        TRUE if null diagonal, otherwise FALSE
      (SIDE)EFFECTS: none
*/
  long int   i;
  
  for ( i = 0 ; i < size ; i++ ) {
    if( matrix[i][i] != 0 ) {
      return FALSE;
    }
  }
  return TRUE;
}

long int check_symmetry ( long int **matrix, long int size ) {
/* 
      FUNCTION:      check whether the Matrix matrix is symmetric
      INPUT:         pointer to the matrix
      OUTPUT:        TRUE if symmetric, otherwise FALSE
      (SIDE)EFFECTS: none
*/
  long int   i, j;
  
  for ( i = 0 ; i < size - 1 ; i++ ) {
    for ( j = i + 1 ; j < size ; j++ ) {
      if( matrix[i][j] != matrix[j][i] )
	return FALSE;
    }
  }
  return TRUE;
}
       
void make_matrix_symmetric( long int **matrix, long int size ) {
/* 
      FUNCTION:      makes an asymmetric matrix symmetric (calculates M = M + M-transpose)
      INPUT:         pointer to the matrix
      OUTPUT:        none
      (SIDE)EFFECTS: makes the Matrix matrix symmetric 
*/
  long int  i, j;  /* index variables */
  long int  help;
  
  for ( i = 0 ; i < size ; i++ ) {
    for ( j = 0 ; j < i ; j++ ) {
      help = matrix[i][j] + matrix[j][i];
      matrix[i][j] = help;
      matrix[j][i] = help;
    }
  } 
}

void print_solution( long int *p ) {
/*    
      FUNCTION:      prints solution p
      INPUT:         pointer to the solution
      OUTPUT:        none
      (SIDE)EFFECTS: none
*/
  int i;

  printf("Assignment: \n");
  for ( i = 0 ; i < n ; i++ ) {
    printf(" %ld ",p[i]);
  }
  printf("\n");
}

void print_matrix ( long int **m ) {
/*    
      FUNCTION:      prints a matrix m
      INPUT:         pointer to the matrix 
      OUTPUT:        none
      (SIDE)EFFECTS: none
*/
  long int i, j;

  printf("\n");
  for ( i = 0 ; i < n ; i++ ) {
    for ( j = 0 ; j < n ; j++ ) {
      printf(" %ld ", m[i][j]);
    }
    printf("\n");
  }
}    

void print_header ( ) {
/*    
      FUNCTION:      prints the header of the output file
      INPUT:         none
      OUTPUT:        none
      (SIDE)EFFECTS: none
*/

  fprintf(output_file,"Machine Pentium III, 700 MHz, Cache unknown\n");
  fprintf(output_file,"Compiler gcc  egcs-2.91.66 19990314 Flags -O -ansi -Wall\n");
  fprintf(output_file,"OS SuSe Linux 6.1\n\n");
  fprintf(output_file,"Algorithm Robust Tabu Search for QAPs, V1.0\n\n");
  fprintf(output_file,"Parameter Settings\n\n");
  fprintf(output_file,"max-tries %ld\n", max_trials);
  fprintf(output_file,"tabu-length %ld\n", tabu_search_length);
  fprintf(output_file,"time %f\n\n", time_limit);

  fprintf(output_file,"Initialization time %f\n\n", elapsed_time ( VIRTUAL ) );

  fprintf(output_file,"begin problem %s\n",name_buf);
}

long int compute_evaluation_function( long int *p ) {
/*    
      FUNCTION:      computes the objective function value for the QAP
      INPUT:         pointer to a solution 
      OUTPUT:        none
      (SIDE)EFFECTS: none
      COMMENTS:      Division by 2 has to be done if we have a n asymmetric instance which 
                     has been converted into a symmetric one (indicated by make_symmetric_flag). 
		     This is due to the particular way of doing this conversion.
*/
   long int  i, j;
   unsigned long  obj_f_value; /* unsigned, because with only "long int" we have an overflow
				  on some few instances, for example, tai100b. This is because 
				  of making this instance symmetric (see make_matrix_symmetric) 
			       */
   obj_f_value = 0;
   for ( i = 0 ; i < n ; i++ ) {
     for ( j = 0 ; j < n ; j++ ) {
       obj_f_value += d[i][j] * f[p[i]][p[j]];
     }
   }
   if ( make_symmetric_flag ) 
     obj_f_value /= 2;
   VERYVERBOSE ( printf("objective function value = %lu \n\n",obj_f_value); )
   return obj_f_value;
}

long int * generate_random_vector( long int size )
/*    
      FUNCTION:      generates a random vector, quick and dirty
      INPUT:         vector dimension
      OUTPUT:        returns pointer to vector, free memory after using the vector
      (SIDE)EFFECTS: none
*/
{
   long int  i, j, help;
   long int  *v;

   v = malloc( size * sizeof(long int) );

   for ( i = 0 ; i < size; i++ ) 
     v[i] = i;

   for ( i = 0 ; i < size-1 ; i++) {
     j = (long int) ( ran01( &seed ) * (size - i)); 
     assert( i + j < size );
     help = v[i];
     v[i] = v[i+j];
     v[i+j] = help;
   }
   VERYVERBOSE ( printf("Random vector:\n");
   for (i = 0 ; i < size ; i++ ) 
     printf(" %ld ",v[i]);
   printf("\n"); )
   return v;
}

void best_2_opt_asymmetric ( long int * q ) {
/*    
      FUNCTION:      best improvement 2-opt local search for symmetric instances
      INPUT:         pointer to some initial assignment
      OUTPUT:        none
      (SIDE)EFFECTS: initial assignment is locally optimized
      COMMENTS:      fast evaluation of moves with additional matrix move_values
                     the first local search iteration is slow 
                     local search looks for best neighboring solution in each iteration
*/

  long int   u, v, k;
  long int   r, s;                      /* memorize which is best move in previous iteration */
  long int   tmp;
  long int   original_symmetric_factor; /* = 2: original symmetric instance
					   = 1: original asymmetric instance 
					*/

  VERYVERBOSE ( printf("best imp, asymmetric case\n"); )
  if ( make_symmetric_flag )
    original_symmetric_factor = 1;
  else
    original_symmetric_factor = 2;

  r = rchosen;
  s = schosen;

  /* in the first local search iteration the full neighborhood has to be evaluated */
  if (first_it_flag) {
    first_it_flag = FALSE;
    for ( u = 0 ; u < n-1 ; u++) {
      for ( v = u+1 ; v < n ; v++) {
	tmp = 0;
	for ( k = 0 ; k < n ; k++ ) {
	  if ( (k != u) && (k != v) ) {
	    tmp += d[k][u] * ( f[q[k]][q[v]] - f[q[k]][q[u]] ) + 
	      d[k][v] * ( f[q[k]][q[u]] - f[q[k]][q[v]] ) + 
	      d[u][k] * ( f[q[v]][q[k]] - f[q[u]][q[k]] ) + 
	      d[v][k] * ( f[q[u]][q[k]] - f[q[v]][q[k]] );
	  }    
	}
	tmp += d[u][u] * ( f[q[v]][q[v]] - f[q[u]][q[u]] )+
	  d[u][v] * ( f[q[v]][q[u]] - f[q[u]][q[v]] )+
	  d[v][u] * ( f[q[u]][q[v]] - f[q[v]][q[u]] )+ 
	  d[v][v] * ( f[q[u]][q[u]] - f[q[v]][q[v]] );
	move_values[u][v] = tmp;
      }
    }
  } else {
    for ( u = 0 ; u < n-1 ; u++) {
      for ( v = u+1 ; v < n ; v++) {
	if (u == r || v == s || u == s || v == r) {
	  tmp = 0;
	  for ( k = 0 ; k < n ; k++ ) {
	    if ( (k != u) && (k != v) ) {
	      tmp += d[k][u] * ( f[q[k]][q[v]] - f[q[k]][q[u]] ) + 
		d[k][v] * ( f[q[k]][q[u]] - f[q[k]][q[v]] ) + 
		d[u][k] * ( f[q[v]][q[k]] - f[q[u]][q[k]] ) + 
		d[v][k] * ( f[q[u]][q[k]] - f[q[v]][q[k]] );
	    }    
	  }
	  tmp += d[u][u] * ( f[q[v]][q[v]] - f[q[u]][q[u]] )+
	    d[u][v] * ( f[q[v]][q[u]] - f[q[u]][q[v]] )+
	    d[v][u] * ( f[q[u]][q[v]] - f[q[v]][q[u]] )+ 
	    d[v][v] * ( f[q[u]][q[u]] - f[q[v]][q[v]] );
	  move_values[u][v] = tmp;
	} else { /* change derived from move_values */
	  tmp = ( d[r][u] - d[r][v] + d[s][v] - d[s][u] ) *
	    ( f[q[s]][q[u]] - f[q[s]][q[v]] + f[q[r]][q[v]] - f[q[r]][q[u]] )
	      + ( d[u][r] - d[v][r] + d[v][s] - d[u][s] ) *
	    ( f[q[u]][q[s]] - f[q[v]][q[s]] + f[q[v]][q[r]] - f[q[u]][q[r]] );
	  tmp += move_values[u][v];
	  move_values[u][v] = tmp;
	}
      }
    }
  }
}

void best_2_opt_symmetric ( long int * q ) {
/*    
      FUNCTION:      best improvement 2-opt local search for symmetric instances
      INPUT:         pointer to some initial assignment
      OUTPUT:        none
      (SIDE)EFFECTS: initial assignment is locally optimized
      COMMENTS:      faster evaluation of moves with additional matrix move_values
*/

  long int   u, v, k;
  long int   r, s;
  long int   tmp;
  long int   original_symmetric_factor;

  VERYVERBOSE ( printf("best imp, symmetric case\n"); )
  if ( make_symmetric_flag )
    original_symmetric_factor = 1;
  else
    original_symmetric_factor = 2;

  r = rchosen;
  s = schosen;

  /* in the first iteration the full neighborhood has to be evaluated */
  if (first_it_flag) {
    first_it_flag = FALSE;
    for ( u = 0 ; u < n-1 ; u++) {
      for ( v = u+1 ; v < n ; v++) {
	tmp = 0;
	for ( k = 0 ; k < n ; k++ ) {
	  if ( (k != u) && (k != v) ) {
	    tmp += ( d[k][u] - d[k][v] ) * ( f[q[k]][q[v]] - f[q[k]][q[u]] );
	  }    
	}
	move_values[u][v] = tmp * original_symmetric_factor;
      }
    }
  } else {
    for ( u = 0 ; u < n-1 ; u++) {
      for ( v = u+1 ; v < n ; v++) {
	if (u == r || v == s || u == s || v == r) {
	  tmp = 0.;
	  for (k = 0 ; k < n ; k++) {
	    if ( (k != u) && (k != v) ) {
	      tmp += ( d[k][u] - d[k][v] ) * ( f[q[k]][q[v]] - f[q[k]][q[u]] );
	    }    
	  }
	  move_values[u][v] = tmp * original_symmetric_factor;
	} else { /* change derived from prev iteration, u and v differ from rchosen or schosen */
	  tmp = original_symmetric_factor * ( d[r][u] - d[r][v] + d[s][v] - d[s][u] ) *
	    ( f[q[s]][q[u]] - f[q[s]][q[v]] + f[q[r]][q[v]] - f[q[r]][q[u]] );
	  tmp += move_values[u][v];
	  move_values[u][v] = tmp;
	}
      }
    }
  }
}

long int choose_tabu_length( )
/*    
      FUNCTION:      choose a new tabu list length
      INPUT:         instance size
      OUTPUT:        none
      (SIDE)EFFECTS: changes the value of variable tabu__list_length
*/
{
  double  help, min, max;

  min      = 0.9  * n;
  max      = 1.1  * n;
  help     = min + ran01( &seed ) * (max - min);
  return   MAX(2,((long int) help));
}

long int ** init_tabu( )
/*    
      FUNCTION:      allocate and initialize the tabu values
      INPUT:         instance size
      OUTPUT:        pointer to matrix, has to be freed before next trial
      (SIDE)EFFECTS: current assignment is modified
*/
{
  long int i, j;
  long int     **matrix;

  VERYVERBOSE ( printf("initialze matrix of tabu values\n"); ) 

  if((matrix = malloc(sizeof(long int) * n * n +
		      sizeof(long int *) * n )) == NULL){
    printf("Out of memory, exit.");
    exit(1);
  }

  for ( i = 0 ; i < n ; i++ ) {
    matrix[i] = (long int *)(matrix + n) + i*n;
    for (j = 0 ; j < n ; j++ ) {
      matrix[i][j] = 0;
    }
  }
  return matrix;
}

long int ** init_move_values( )
/*    
      FUNCTION:      allocate and initialize the tabu values
      INPUT:         instance size
      OUTPUT:        pointer to matrix, has to be freed before next trial
      (SIDE)EFFECTS: current assignment is modified
*/
{
  long int i, j;
  long int     **matrix;

  VERYVERBOSE ( printf("initialze matrix of move values\n"); ) 

  if((matrix = malloc(sizeof(long int) * n * n +
		      sizeof(long int *) * n	 )) == NULL){
    printf("Out of memory, exit.");
    exit(1);
  }

  for ( i = 0 ; i < n ; i++ ) {
    matrix[i] = (long int *)(matrix + n) + i*n;
    for (j = 0 ; j < n ; j++ ) {
      matrix[i][j] = 0;
    }
  }
  return matrix;
}

void swap( long int pos_1, long int pos_2, long int *q ) {   
/*    
      FUNCTION:      swap items at positions pos_1 and pos_2
      INPUT:         positions 1 and 2, pointer to current assignment
      OUTPUT:        none
      (SIDE)EFFECTS: current assignment is modified
*/
  long int  help;
  
  help     = q[pos_1];
  q[pos_1] = q[pos_2];
  q[pos_2] = help;
}

void  make_tabu( long int * q, long int iteration, long int r, long int s )
/*    
      FUNCTION:      make an invers move tabu (forbids a location for an item)
      INPUT:         pointer to some assignment, iteration number, two locations involved in move
      OUTPUT:        none
      (SIDE)EFFECTS: matrix of tabu_values is modified
*/
{
  tabu_values[r][q[r]] = iteration + tabu_list_length;
  tabu_values[s][q[s]] = iteration + tabu_list_length;
}

void select_move_taillard( long int *q, long int current, long int iteration )
/*    
      FUNCTION:      select the best move which is not tabu or is aspired
      INPUT:         pointer to some assignment, obj_f_val of current solution, iteration number
      OUTPUT:        none
      (SIDE)EFFECTS: global variables rchosen and schosen are assigned the locations involved in the move
      COMMENT:       this adapts the selection criterion to the one proposed in the recent code of Taillard: of the forced moves take the best one if several are available
*/
{
  long int   i, j;
  long int   max_decrease;
  long int   taboo, aspired, already_aspired;
  
  max_decrease = LONG_MAX;
  already_aspired = FALSE;
  rchosen = n; schosen = n;
  for ( i = 0 ; i < n - 1 ; i++) {
    for (j = (i+1); j < n ; j++) {
      if ( (tabu_values[i][q[j]] > iteration ) && (tabu_values[j][q[i]] > iteration ) )
	taboo = TRUE;
      else 
	taboo = FALSE;
      if ( ((current + move_values[i][j]) < best_found ) && taboo )
	aspired = TRUE;
      else if ( tabu_values[i][q[j]] < aspirating_iteration || 
		tabu_values[j][q[i]] < aspirating_iteration ) {
	aspired = TRUE;
/* 	printf("iteration %ld, aspired forced ",iteration); */
      }
      else 
	aspired = FALSE;
/*       if ( (move_values[i][j] < max_decrease && !taboo) || aspired ) { */
      if ( (aspired && !already_aspired) || 
	   (aspired && already_aspired && move_values[i][j] < max_decrease ) || 
	   (!aspired && !already_aspired && move_values[i][j] < max_decrease && !taboo)) {
	rchosen = i;
	schosen = j;
	if (aspired)
	   already_aspired = TRUE;
	/* 	if ( aspired && !taboo ) {
	  max_decrease = LONG_MIN;
	}
	else */
	max_decrease = move_values[i][j];
      }
    }
  }
  assert ( rchosen >= 0 && rchosen < n);
  assert ( schosen >= 0 && schosen < n);
/*   if (  max_decrease == LONG_MIN ) */
/*     printf("\n"); */
}


void select_move( long int *q, long int current, long int iteration )
/*    
      FUNCTION:      select the best move which is not tabu or is aspired
      INPUT:         pointer to some assignment, obj_f_val of current solution, iteration number
      OUTPUT:        none
      (SIDE)EFFECTS: global variables rchosen and schosen are assigned the locations involved in the move
      COMMENT:       this function implements the selection rule without considering that several
                     forced moves are executed; here, the first one detected is executed.
		     Maybe test with another one, where a random forced move is executed.
*/
{
  long int   i, j;
  long int   max_decrease;
  long int   taboo, aspired;
  
  max_decrease = LONG_MAX;
  rchosen = n; schosen = n;
  for ( i = 0 ; i < n - 1 ; i++) {
    for (j = (i+1); j < n ; j++) {
      if ( (tabu_values[i][q[j]] > iteration ) && (tabu_values[j][q[i]] > iteration ) )
	taboo = TRUE;
      else 
	taboo = FALSE;
      if ( ((current + move_values[i][j]) < best_found ) && taboo )
	aspired = TRUE;
      else if ( tabu_values[i][q[j]] < aspirating_iteration && 
		tabu_values[j][q[i]] < aspirating_iteration ) {
	aspired = TRUE;
/* 	printf("iteration %ld, aspired forced ",iteration); */
      }
      else 
	aspired = FALSE;
      if ( (move_values[i][j] < max_decrease && !taboo) || aspired ) {
	rchosen = i;
	schosen = j;
 	if ( aspired && !taboo ) {
	  max_decrease = LONG_MIN;
	}
	else
	  max_decrease = move_values[i][j];
      }
    }
  }
  assert ( rchosen >= 0 && rchosen < n);
  assert ( schosen >= 0 && schosen < n);
/*   if (  max_decrease == LONG_MIN ) */
/*     printf("\n"); */
}


void tabu_search( )
/*    
      FUNCTION:      perform robust tabu search
      INPUT:         pointer to initial solution, instance size
      OUTPUT:        pointer to best solution
      (SIDE)EFFECTS: 
*/
{
  long int  i;
  long int  iterations, obj_f_value, *b, *q;

  VERYVERBOSE ( printf("tabu search\n"); )
  move_values = init_move_values( );
  tabu_values = init_tabu( );
  b = generate_random_vector( n );
  q = malloc( n * sizeof(long int) );
  for ( i = 0 ; i < n ; i++ ) 
    q[i] = b[i];
  obj_f_value       = compute_evaluation_function( b );
  best_found        = obj_f_value;
  first_it_flag     = TRUE;  
  rchosen           = schosen      = 0;
  init_tabu();
  iterations        = 1;
  tabu_list_length  = choose_tabu_length();

  while ( iterations < tabu_search_length || elapsed_time( VIRTUAL ) < time_limit ) {

    aspirating_iteration = iterations - iterations_before_aspiration;

    if ( d_symmetric_flag && f_symmetric_flag && null_diagonal_flag )
      best_2_opt_symmetric ( q );
    else if ( make_symmetric_flag )
      best_2_opt_symmetric( q );
    else
      best_2_opt_asymmetric( q );

    select_move( q , obj_f_value, iterations );

    make_tabu( q, iterations, rchosen, schosen ); /* make_tabu has to go before swap */

    swap( rchosen, schosen, q );

    obj_f_value += move_values[rchosen][schosen];

    if ( obj_f_value < best_found ) {
      best_found = obj_f_value;
      printf("Best so far %ld\t iterations %ld\t time %f\n",best_found,iterations,elapsed_time( VIRTUAL ));
      fprintf(output_file,"best %ld\tcycle %ld\tsteps %ld\ttime %f\n",best_found,iterations,iterations,elapsed_time( VIRTUAL ));
      time_best_found = elapsed_time( VIRTUAL );
      iteration_best_found = iterations;
      for ( i = 0 ; i < n ; i++ ) 
	b[i] = q[i];
    }

    if ( !(iterations % ( long int )(2.2 * n + 4))) {
      tabu_list_length = choose_tabu_length();
      VERYVERBOSE ( printf(" tabu_length = %ld, iteration %ld\n",tabu_list_length, iterations); )
printf(" tabu_length = %ld, iteration %ld\n",tabu_list_length, iterations);
    }
    iterations++;

  }
  print_solution ( b );
  fprintf(output_summary_file,"best %ld\titeration %ld\ttime to best %.2f\ttotal time %.2f\n",best_found, iteration_best_found, time_best_found,elapsed_time( VIRTUAL ));

  if ( compute_evaluation_function( b ) != best_found )
    fprintf(stderr,"Some error must have occurred in local search routine,\n values do not match\n");
  free ( b );
  free ( q );
  free ( move_values );
  free ( tabu_values );

}

void usage()
/*    
      FUNCTION:      output usage of program
      INPUT:         none
      OUTPUT:        none
      (SIDE)EFFECTS: 
*/
{
  fprintf(stderr,"\nRobust Tabu Search, V1.0\n");
  fprintf(stderr,"usage: [- see below -] [-i dat_file] (given are long and short options)\n");
  fprintf(stderr,"--trials     -r #\t number of trials to be run on one instance\n");
  fprintf(stderr,"--time       -t #\t maximum time limit (not used for the moment)\n");
  fprintf(stderr,"--tabu       -l #\t length of one tabu search run as multiple of instance size\n");
  fprintf(stderr,"--input      -i #\t input file\n");
  fprintf(stderr,"--help       -h \t\t help: prints this information\n");

}

void set_default_parameters()
/*    
      FUNCTION:      set default parameter values
      INPUT:         none
      OUTPUT:        none
      (SIDE)EFFECTS: assigns some values to global variables
*/
{
  max_trials         = 10;
  tabu_search_length = 10000;
  time_limit         = 0.0;

}

void checkOutOfRange(int value, int MIN, int MAX, char *optionName){
  if ((value<MIN)||(value>MAX)){
    fprintf(stderr,"Error: Option `%s' out of range\n",optionName);
    exit(1);
  }
}

void check_options(int argc, char *argv[]) {
  
  while (1){
    int option_index = 0;

    static struct option long_options[] = {
      {"time",        1, 0, 't'},
      {"trials",      1, 0, 'r'},
      {"tabu",        1, 0, 'l'},
      {"input",       1, 0, 'i'},
      {"help",        0, 0, 'h'},
      {0,             0, 0,  0 }
    };
    
    opt = getopt_long(argc, argv, "t:r:l:i:h",
		       long_options, &option_index);
    if (opt == -1)
      break;
    
    switch (opt){
    case 0:
      fprintf(stderr,"Error: Confused by input on long option...\n"), exit(1);
      break;
      
    case 't':
      time_limit = atof(optarg);
      break;
      
    case 'r':
      max_trials = atol(optarg);
      checkOutOfRange(max_trials,1,1000,"r (trials)");
      break;
      
    case 'l':
      tabu_search_length = atol(optarg);
      checkOutOfRange(tabu_search_length,0,100000,"l (tabu_search_length)");
      break;

    case 'i':
      strncpy(name_buf,optarg,strlen(optarg));
      break;

    case 'h':
      usage();
      fprintf(stderr,"No more help yet, sorry.\n"), exit(1);
      break;
      
    default:
      fprintf(stderr,"Error: Confused by input... %ld\n",opt), exit(1);
      
    }
  }
}  

void init_program( long int argc, char *argv[] )
/*    
      FUNCTION:      all the stuff for starting the program: input, parameters, etc.
      INPUT:         argument list from program call
      OUTPUT:        none
      (SIDE)EFFECTS: 
*/
{
  FILE *qap_file;
  long int i;
  char buf[LINE_BUF_LEN], *c;    /* input file name */

  setbuf(stdout,NULL);
  printf(PROG_ID_STR);
  seed = (long int) time(NULL); /* initialize random number generator */

  if( (qap_file = fopen(name_buf, "r")) == NULL) {
    fprintf(stderr, "error opening input file %s\n",name_buf);
    exit(1);	
  }
  
  read_problem_size( qap_file );
  
  read_best_known_value( qap_file );
  
  d = read_matrix( qap_file, n);
  d_symmetric_flag = check_symmetry ( d, n );
  null_diagonal_flag = check_null_diagonal ( d, n );   
  VERYVERBOSE ( print_matrix( d ); )
    
  f = read_matrix( qap_file, n);
  f_symmetric_flag = check_symmetry ( f, n );
  if ( null_diagonal_flag )
    ; /* if one matrix has already null diagonal we need not check the other */
  else
    null_diagonal_flag = check_null_diagonal ( f, n );
  VERYVERBOSE ( print_matrix( f ); )
    
  VERBOSE ( printf("d_symmetric_flag %ld, f_symmetric_flag %ld, null_diagonal_flag %ld\n",d_symmetric_flag, f_symmetric_flag, null_diagonal_flag); )

  make_symmetric_flag = XOR(d_symmetric_flag,f_symmetric_flag);
  if ( make_symmetric_flag && null_diagonal_flag ) {
    if ( !d_symmetric_flag )
      make_matrix_symmetric ( d, n );
    else if ( !f_symmetric_flag )
      make_matrix_symmetric ( f, n );
    else {
      fprintf(stderr,"One matrix should have been symmetric\n");
      exit(1);
    }
  }
   
  if ( ( c = strrchr(name_buf,'/') ) != NULL ) {
    c++;
    i = 0;
    while ((buf[i] = *c) != '\0') {
      i++;
      c++;
    }
    printf("%s\n",buf);
    strcpy(name_buf, buf);
  }

  sprintf(buf,"%s.RoTS.v.1.0.rep",name_buf);

  if( (output_file = fopen(buf, "w")) == NULL) {
    fprintf(stderr, "error opening output file %s\n","qap.rep");
    exit(1);	
  }

  sprintf(buf,"%s.RoTS.v.1.0.sum",name_buf);
  
  if( (output_summary_file = fopen(buf, "w")) == NULL) {
     fprintf(stderr, "error opening output file %s\n","qap.sum");
     exit(1);	
  }
 
  /* finally set these parameters to the right value, they were defined as multiples of instance size */
  tabu_search_length *= n;
  iterations_before_aspiration = (long int)(3. * n * n);

  best_in_try = malloc( max_trials * sizeof(long int) );
  best_found_at = malloc( max_trials * sizeof(long int) );
  t_best_found = malloc( max_trials * sizeof(double) );
  
}

void start_trial( )
{
  start_timers();
  fprintf(output_file,"begin try %li\n",trials);
  VERBOSE ( printf("begin try %li\n",trials); )
}

void end_trial( )
{
  fprintf(output_file,"end try %ld\t total time %f\n",trials, elapsed_time( VIRTUAL ));
  VERBOSE ( printf("end try %li\n",trials); )
  best_in_try[trials-1] = best_found;
  best_found_at[trials-1] = iteration_best_found;
  t_best_found[trials-1] = time_best_found;
}

void end_program( )
{
  fprintf(output_file,"end problem %s\n",name_buf);
  VERBOSE ( printf("end problem %s\n",name_buf); )
}

void exit_program( void ) 
{
  long int best_tour_length, worst_tour_length;
  double   t_avgbest, t_stdbest;
  double   avg_sol_quality = 0., avg_cyc_to_bst = 0., stddev_best, stddev_iterations;

  fprintf(output_summary_file,"\n\nMax tries %ld",max_trials);

  best_tour_length = best_of_vector( best_in_try , max_trials );
  worst_tour_length = worst_of_vector( best_in_try , max_trials );
  avg_cyc_to_bst = mean( best_found_at , max_trials );
  stddev_iterations = std_deviation( best_found_at, max_trials, avg_cyc_to_bst );
  avg_sol_quality = mean( best_in_try ,max_trials );
  stddev_best = std_deviation( best_in_try, max_trials, avg_sol_quality);
  t_avgbest = meanr( t_best_found, max_trials );
  printf(" t_avgbest = %f\n", t_avgbest );
  t_stdbest = std_deviationr( t_best_found, max_trials, t_avgbest);
 
  fprintf(output_summary_file,"\nAverage-Best: %.2f\t Average-Iterations: %.2f", avg_sol_quality, avg_cyc_to_bst);
  fprintf(output_summary_file,"\nStddev-Best: %.2f \t Stddev Iterations: %.2f", stddev_best, stddev_iterations);
  fprintf(output_summary_file,"\nBest try: %ld \n", best_tour_length);
  fprintf(output_summary_file,"\nAvg.time-best: %.2f stddev.time-best: %.2f\n", t_avgbest, t_stdbest); 

  if ( best_known > 0 ) {
    fprintf(output_summary_file," excess best = %f, excess average = %f, excess worst = %f\n",(double)(best_tour_length - best_known) / (double)best_known,(double)(avg_sol_quality - best_known) / (double)best_known,(double)(worst_tour_length - best_known) / (double)best_known);
  }
}

/* --- main program ------------------------------------------------------ */

int main(int argc, char *argv[]) {
  

  /* variables for main programs are declared above */
  

  start_timers();             /* start timing routines */

  set_default_parameters();
  
  check_options(argc, argv);
  
  init_program(argc, argv);   /* initialize all important data */

  print_header ();

  for ( trials = 1 ; trials <= max_trials ; trials ++ ) {

    start_trial();

    tabu_search();

    end_trial();
    fflush ( output_file );
    fflush ( output_summary_file );

  }

  fprintf(output_file,"end problem %s\n",name_buf);
  

  VERBOSE ( printf("best found %ld, percentage deviation from best known %f, run-time %f\n",best_found,((float)best_found - (float)best_known) / (float)best_known * 100., elapsed_time ( VIRTUAL )); )

  exit_program();

	
  fflush ( output_file );
  fflush ( output_summary_file );
  
  free ( d );  /* first matrix */
  free ( f );  /* second matrix */
  free ( p );  /* solution vector */
  free ( best_in_try );
  free ( best_found_at );
  free ( t_best_found );
  return(0);
}











